"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.StringFormat = void 0;
const assert_util_1 = require("./util/assert-util");
/**
 * String Format class.
 *
 * The functionality supports python-style string formatting, and C-style string formatting.
 *
 * @class StringFormat
 * @author RubixCode
 * @since 0.1.0
 */
class StringFormat {
    /**
     * set of formatters which can be added or removed
     */
    formatters = new Map();
    /**
     * use multiple formatters
     */
    multipleFormatterMode = false;
    /**
     * Enable Multiple Formatter mode.
     * This will cause the string to be formatted by all the string formatters.
     */
    enableMultipleFormatterMode() {
        this.multipleFormatterMode = true;
    }
    /**
     * Disable Multiple Formatter mode.
     * This will cause the string to be formatted by the first string formatter that can parse it.
     * If no formatter can parse the string, the string will be returned as is.
     */
    disableMultipleFormatterMode() {
        this.multipleFormatterMode = false;
    }
    /**
     * is multiple formatter enabled
     */
    isMultipleFormatterModeEnabled() {
        return this.multipleFormatterMode;
    }
    /**
     * Add a formatter to the set of formatters.
     *
     * @param {IStringFormatter} formatter the formatter
     * @since 0.1.0
     */
    addFormatter(formatter) {
        const identifier = formatter.name;
        if (this.formatters.has(identifier)) {
            throw new Error(`Formatter ${identifier} already exists`);
        }
        this.formatters.set(identifier, formatter);
        return this;
    }
    /**
     * Check if the formatter exists
     *
     * @param {string} identifier the formatter name
     * @returns {boolean} true if the formatter exists
     * @since 0.1.0
     */
    hasFormatter(identifier) {
        return this.formatters.has(identifier);
    }
    /**
     * Remove a formatter from the set of formatters.
     *
     * @param {string} identifier the identifier of the formatter
     * @since 0.1.0
     */
    removeFormatter(identifier) {
        (0, assert_util_1.isString)("identifier", identifier);
        if (!this.formatters.has(identifier)) {
            throw new Error(`Formatter '${identifier}' does not exist`);
        }
        // remove the formatter from the set
        this.formatters.delete(identifier);
    }
    constructor() { }
    /**
     * Adds ".format" function to the object.
     * @param {Class} obj Object
     */
    extend(obj) {
        const self = this;
        // register prototype method
        obj.prototype.format = function (...args) {
            return self.stringFormat(this.toString(), ...args);
        };
    }
    /**
     * Format a string
     *
     * @param {string} input the input string
     * @param {any[]} param the parameters
     * @returns the formatted string
     */
    format(input, ...param) {
        return this.stringFormat(input, ...param);
    }
    /**
     * Format a string.
     *
     * @param {string} line the line to format
     * @param {any[]} params the parameters to format the line with
     * @returns the formatted line
     */
    stringFormat(line, ...params) {
        let result = line;
        // iterate over the formatters
        for (const formatter of this.formatters.values()) {
            // try to format the line
            const [isFormatted, formattedLine,] = formatter.format(result, ...params);
            if (isFormatted) {
                result = formattedLine;
                if (!this.multipleFormatterMode) {
                    break;
                }
            }
        }
        return result;
    }
}
exports.StringFormat = StringFormat;
