"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const assert_util_1 = require("../util/assert-util");
const TypeErrors = {
    INTEGER_TYPE_ERROR(value) { return new TypeError(`${value} is not an integer`); },
    STRING_TYPE_ERROR(value) { return new TypeError(`${value} is not a string`); },
};
/**
 * function to check if string can be converted to number
 * @param value number in string
 * @returns boolean
 */
function isParsableNumber(value) {
    let n = parseFloat(value);
    let result = !isNaN(n) && isFinite(n);
    return result;
}
//#region String format parsers - C style
//%d %s - simple integer and string
const stringIntParser = {
    name: "string-int-parser",
    testRegex: /^([ds])$/,
    formatter(identifier, line, param) {
        if (RegExp.$1 === "d" && !isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        else if (RegExp.$1 === "s" && typeof param !== "string" && typeof param !== "boolean") {
            throw TypeErrors.STRING_TYPE_ERROR(param);
        }
        return [
            line.substring(0, line.indexOf("%" + identifier)),
            param,
            line.substring(line.indexOf("%" + identifier) + identifier.length + 1),
        ].join("");
    },
};
// %5d - padded integer
const paddedIntParser = {
    name: "padded-int-parser",
    testRegex: /^([0\-]?)(\d+)d$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        let length = parseInt(RegExp.$2) - param.toString().length;
        let regex2int = parseInt(RegExp.$2);
        let replaceStr = "";
        if (length < 0) {
            length = 0;
        }
        // Won't have a default case due to the regex
        // tslint:disable-next-line: switch-default
        switch (RegExp.$1) {
            case "": // right align
                replaceStr = (Array(length + 1).join(" ") + param).slice(-regex2int);
                break;
            case "-": // left align
                replaceStr = (param + Array(length + 1).join(" ")).slice(-regex2int);
                break;
            case "0": // zero fill
                replaceStr = (Array(length + 1).join("0") + param).slice(-regex2int);
                break;
        }
        return line.replace("%" + identifier, replaceStr);
    },
};
// %o - octet
const octetParser = {
    name: "octet-parser",
    testRegex: /^([o])$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        return line.replace("%" + identifier, parseInt(param).toString(8));
    },
};
// %b - binary
const binaryParser = {
    name: "binary-parser",
    testRegex: /^(b)$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        return line.replace("%" + identifier, parseInt(param).toString(2));
    },
};
// %x, %X - hexadecimal
const hexParser = {
    name: "hex-parser",
    testRegex: /^([xX])$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        let hex = parseInt(param).toString(16);
        if (identifier === "X") {
            hex = hex.toUpperCase();
        }
        return line.replace("%" + identifier, hex);
    },
};
// %c - character
const charParser = {
    name: "char-parser",
    testRegex: /^(c)$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        return line.replace("%" + identifier, String.fromCharCode(parseInt(param)));
    },
};
// %5s - padded string
const paddedStringParser = {
    name: "padded-string-parser",
    testRegex: /^(-?)(\d+)s$/,
    formatter(identifier, line, param) {
        if (typeof param !== "string") {
            throw TypeErrors.STRING_TYPE_ERROR(param);
        }
        let length = parseInt(RegExp.$2) - param.toString().length, regex2int = parseInt(RegExp.$2), replaceStr = "";
        if (length < 0) {
            length = 0;
        }
        // Won't have a default case due to the regex
        // tslint:disable-next-line: switch-default
        switch (RegExp.$1) {
            case "": // right align
                replaceStr = (Array(length + 1).join(" ") + param).slice(-regex2int);
                break;
            case "-": // left align
                replaceStr = (param + Array(length + 1).join(" ")).slice(-regex2int);
                break;
        }
        return [
            line.substring(0, line.indexOf("%" + identifier)),
            replaceStr,
            line.substring(line.indexOf("%" + identifier) + identifier.length + 1),
        ].join("");
    },
};
// %4.5s - string with max length
const stringMaxParser = {
    name: "string-max-parser",
    testRegex: /^(-?\d*)\.(\d+)s$/,
    formatter(identifier, line, param) {
        if (typeof param !== "string") {
            throw TypeErrors.STRING_TYPE_ERROR(param);
        }
        let max = 0, replaceStr = "", spaceWidth = 0, regex2int = parseInt(RegExp.$2);
        // %.4s
        if (RegExp.$1 === "") {
            replaceStr = param.slice(0, regex2int);
        }
        // %-5.4s or %5.4s
        else {
            param = param.slice(0, regex2int);
            max = Math.abs(parseInt(RegExp.$1));
            spaceWidth = max - param.length;
            replaceStr = RegExp.$1.indexOf("-") !== -1 ?
                (param + Array(spaceWidth + 1).join(" ")).slice(-max) :
                (Array(spaceWidth + 1).join(" ") + param).slice(-max);
        }
        return [
            line.substring(0, line.indexOf("%" + identifier)),
            replaceStr,
            line.substring(line.indexOf("%" + identifier) + identifier.length + 1),
        ].join("");
    },
};
// %u - unicode
const unicodeParser = {
    name: "unicode-parser",
    testRegex: /^(u)$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        return line.replace("%" + identifier, (parseInt(param, 10) >>> 0).toString());
    },
};
// %e - exponential
const exponentialParser = {
    name: "exponential-parser",
    testRegex: /^(-?)(\d*).?(\d?)(e)$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        let leftPad = RegExp.$1 === "-";
        let width = RegExp.$2;
        let decimal = RegExp.$3 !== "" ? Number(RegExp.$3) : undefined;
        let value = Number(param).toExponential(decimal);
        let mantissa, exponent, padLength;
        if (width !== "") {
            if (decimal !== undefined) {
                padLength = parseInt(width) - value.length;
                if (padLength < 0) {
                    // I don't know whats going on here,
                    // and we were not supposed to get here
                    // but in co-pilot we must trust
                    throw new RangeError(`width (${width}) is too small for value (${value})`);
                }
                value = leftPad ?
                    value + Array(padLength + 1).join(" ") :
                    Array(padLength + 1).join(" ") + value;
            }
            else {
                mantissa = value.split("e")[0] || "";
                exponent = "e" + value.split("e")[1];
                padLength = parseInt(width) - mantissa.length - exponent.length;
                value = padLength >= 0 ?
                    mantissa + Array(padLength + 1).join("0") + exponent :
                    mantissa.slice(0, padLength) + exponent;
            }
        }
        return line.replace("%" + identifier, value);
    },
};
// %f - floating point
const floatParser = {
    name: "float-parser",
    testRegex: /^(-?)(\d*).?(\d?)(f)$/,
    formatter(identifier, line, param) {
        if (!isParsableNumber(param)) {
            throw TypeErrors.INTEGER_TYPE_ERROR(param);
        }
        let leftPad = RegExp.$1 === "-", intParam = parseInt(param), width = RegExp.$2, decimal = RegExp.$3, DOT_LENGTH = 1, integerPart = parseInt(param) > 0 ?
            Math.floor(intParam)
            : Math.ceil(intParam), value = parseFloat(param).toFixed(decimal !== "" ? Number(decimal) : 6), mantissaWidth, spaceWidth;
        if (width !== "") {
            if (decimal !== "") {
                mantissaWidth = integerPart.toString().length + DOT_LENGTH + parseInt(decimal, 10);
                spaceWidth = parseInt(width) - mantissaWidth;
                value = leftPad ?
                    parseFloat(param).toFixed(parseInt(decimal)) + Array(spaceWidth + 1).join(" ") :
                    Array(spaceWidth + 1).join(" ") + parseFloat(param).toFixed(parseInt(decimal));
            }
            else {
                value = parseFloat(param).toFixed(parseInt(width) - integerPart.toString().length - DOT_LENGTH);
            }
        }
        return line.replace("%" + identifier, value);
    },
};
//#endregion
/**
 * Support C-Style String formatting
 * The following formats are supported:
 * - %s - string
 * - %d - integer
 * - %u - unicode
 * - %e - exponential
 * - %f - floating point
 * - %c - character
 * - %b - binary
 * - %o - octal
 * - %x - hexadecimal
 * - %X - hexadecimal
 *
 * @see https://github.com/tmaeda1981jp/string-format-js for C-Style formatting
 */
class CStyleFormatter {
    get name() { return this.constructor.name; }
    /**
     * List of formats supported by the C-Style formatter out of the box
     */
    static DEFAULT_FORMAT_PARSERS = [
        stringIntParser,
        paddedIntParser,
        octetParser,
        binaryParser,
        hexParser,
        charParser,
        paddedStringParser,
        stringMaxParser,
        unicodeParser,
        exponentialParser,
        floatParser,
    ];
    /**
     * The regex used to test if a string is a format string
     * This populates RegExp.$1 with the format identifier
     */
    static IDENTIFIER_REGEX = /%([.#0-9\-]*[a-zA-Z])/;
    // set of format parser which can be used to format string
    formatParsers;
    constructor() {
        // add default format parsers
        this.formatParsers = new Map();
        CStyleFormatter.DEFAULT_FORMAT_PARSERS.forEach(parser => {
            this.addFormatParser(parser);
        });
    }
    /**
     * @inheritdoc
     * @since 0.1.0
     */
    isFormatParserSupported() {
        return true;
    }
    /**
     * @inheritdoc
     * @param formatParser the format parser
     * @since 0.1.0
     */
    addFormatParser(formatParser) {
        const identifier = formatParser.name;
        if (this.formatParsers.has(identifier)) {
            throw new Error(`Format parser ${identifier} already exists`);
        }
        this.formatParsers.set(identifier, formatParser);
        // this.logger.debug(`Added format parser ${identifier}`)
    }
    /**
     * @inheritdoc
     * @param {string} identifier the format parser identifier
     * @returns {boolean} true if the format parser is available
     * @since 0.1.0
     */
    hasFormatParser(identifier) {
        return this.formatParsers.has(identifier);
    }
    /**
     * @inheritdoc
     * @param formatParserName the name of the format parser
     * @since 0.1.0
     */
    removeFormatParser(formatParserName) {
        (0, assert_util_1.isString)("formatParserName", formatParserName);
        if (!this.formatParsers.has(formatParserName)) {
            throw new Error(`Format parser ${formatParserName} does not exist`);
        }
        // remove the format parser from the set
        this.formatParsers.delete(formatParserName);
        // this.logger.debug(`Removed format parser ${formatParserName}`)
    }
    /**
     * @inheritdoc
     * @since 0.1.0
     * @returns {false} false transformers feature is not supported by the C-Style formatter
     */
    isTransformersSupported() {
        return false;
    }
    /**
     * @inheritdoc
     * @param key the transformer key
     * @param transformer Transformer
     * @since 0.1.0
     * @throws {Error} if the transformer is not supported
     */
    addTransformer(key, transformer) {
        throw new Error(`Transformers functionality is not supported by the C-Style formatter`);
    }
    /**
     * @inheritdoc
     * @param key the transformer key
     * @returns {boolean} true if the transformer is available
     * @since 0.1.0
     */
    hasTransformer(key) {
        throw new Error(`Transformers functionality is not supported by the C-Style formatter`);
    }
    /**
     * @inheritdoc
     * @param key the transformer key
     * @since 0.1.0
     * @throws {Error} if the transformer is not supported
     */
    removeTransformer(key) {
        throw new Error(`Transformers functionality is not supported by the C-Style formatter`);
    }
    /**
     * @inheritdoc
     * @param {string} input the input string
     * @param {any[]} param the parameters
     * @returns {[boolean, string]} is processed, the formatted string
     * @since 0.1.0
     * @throws {Error} if the input string is not a format string
     */
    format(input, ...params) {
        (0, assert_util_1.isString)("input", input);
        let formatted = false;
        let formattedString = input;
        for (let i = 0; i < params.length; i++) {
            if (formattedString.match(CStyleFormatter.IDENTIFIER_REGEX)) {
                let identifier = RegExp.$1;
                // this.logger.debug(`Found identifier '${identifier}'`)
                // find the formatter
                let formatter;
                for (const f of this.formatParsers.values()) {
                    if (f.testRegex.test(identifier)) {
                        formatter = f;
                        // this.logger.debug(`Found formatter ${formatter.name}`)
                        break;
                    }
                }
                if (formatter) {
                    formattedString = formatter.formatter(identifier, formattedString, params[i]);
                    formatted = true;
                }
            }
        }
        return [formatted, formattedString,];
    }
}
exports.default = CStyleFormatter;
